import sys
import time
from typing import Optional, Dict

import lxml.etree
import requests

from . import config


def flush_mondrian_cache():
    """Reloads the mondrian schema and flushes all caches"""
    url = config.mondrian_server_internal_url() + '/flush-caches'
    try:
        if requests.get(url).status_code == 200:
            return True
    except Exception as e:
        pass

    print(f'Could not call {url}', file=sys.stderr)
    return False



def discover(request_type: str,
             restrictions: Optional[Dict] = None) -> lxml.etree._Element:
    """Creates a discover request and sends it to Mondrian Server.

    Args:
        request_type: Type of the XMLA discover request
        restrictions: Restrictions to place on the request (such as filtering by cube name)

    Returns: The body of the returned response

    """
    restrictions_xmla = "\n".join(
        [f'<{column}>{value}</{column}>' for column, value in restrictions.items()])
    xmla = f"""
    <Discover xmlns="urn:schemas-microsoft-com:xml-analysis">
        <RequestType>{request_type}</RequestType>
        <Restrictions>
            <RestrictionList>
                {restrictions_xmla}
            </RestrictionList>
        </Restrictions>
        <Properties>
           <PropertyList>
              <Catalog>dwh</Catalog>
              <DataSourceInfo>Cubes</DataSourceInfo>
              <Format>Tabular</Format>
              <Content>SchemaData</Content>
           </PropertyList>
        </Properties>
     </Discover>
     """
    return send_xmla_request(xmla)


def execute(query: str) -> lxml.etree._Element:
    """Creates an execute request and sends it to Mondrian Server.

    Args:
        query: MDX query to execute in Mondrian Server

    Returns: The body of the returned response

    """
    xmla = f"""
    <Execute xmlns="urn:schemas-microsoft-com:xml-analysis">
        <Command>
           <Statement>
              <![CDATA[{query}]]>
           </Statement>
        </Command>
        <Properties>
           <PropertyList>
              <Catalog>Mondrian</Catalog>
              <DataSourceInfo>Mondrian</DataSourceInfo>
              <Format>Multidimensional</Format>
              <Content>Data</Content>
              <AxisFormat>TupleFormat</AxisFormat>
           </PropertyList>
        </Properties>
     </Execute>"""
    return send_xmla_request(xmla)


def send_xmla_request(xmla: str, attempt: Optional[int] = 0) -> lxml.etree._Element:
    """Constructs an XMLA request and sends it to Mondrian Server including retries on failed attempts.
    Parses the

    Args:
        xmla: XMLA to be send as the body of the request
        attempt: How many times the request was already attempted

    Returns: Body of the parsed response

    """
    soap_xml = f"""<?xml version="1.0"?>
    <SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
        <SOAP-ENV:Header />
        <SOAP-ENV:Body>{xmla}</SOAP-ENV:Body>
    </SOAP-ENV:Envelope>
    """
    session = requests.Session()
    session.trust_env = False
    response = session.post(f'{config.mondrian_server_internal_url()}/xmla', data=soap_xml.encode('utf-8'),
                            headers={'Content-type': 'text/xml'})

    utf8_parser = lxml.etree.XMLParser(ns_clean=True, encoding='utf-8')
    tree = lxml.etree.fromstring(response.text.encode('utf-8'), parser=utf8_parser)
    body = tree.find("SOAP-ENV:Body", namespaces=tree.nsmap)
    fault = body.find("SOAP-ENV:Fault", namespaces=tree.nsmap)
    if fault is not None:
        if attempt < 1:
            time.sleep(2)
            return send_xmla_request(xmla, attempt=(attempt + 1))
        else:
            code = fault.find('faultcode').text
            error = fault.find('faultstring').text
            description = fault.find(
                'detail'
            ).find(
                "XA:error", namespaces={"XA": "http://mondrian.sourceforge.net"}
            ).find("desc").text

            raise MondrianError(f"""{code} - {error}
            {description}""")
    else:
        return body


class MondrianError(Exception):
    pass



"""Saiku and Mondrian configuration"""

def mondrian_server_internal_url():
    """The url where to reach the mondrian server on the internal network from Mara"""
    return 'http://127.0.0.1:8080'


def mondrian_server_external_url():
    """The url under which end users can reach Saiku"""
    return 'http://127.0.0.1:8080'




"""Make the functionalities of this package auto-discoverable by mara-app"""


def MARA_CONFIG_MODULES():
    from . import config
    return [config]


def MARA_FLASK_BLUEPRINTS():
    from . import views
    return [views.blueprint]


def MARA_ACL_RESOURCES():
    from . import views
    return {
        'Saiku': views.acl_resource_saiku
    }


def MARA_CLICK_COMMANDS():
    return []


def MARA_NAVIGATION_ENTRIES():
    from . import views
    return {
        'Saiku': views.saiku_navigation_entry()
    }



import pathlib
import typing

from lxml import etree
from mara_schema.attribute import Type, Attribute
from mara_schema.data_set import DataSet
from mara_schema.metric import SimpleMetric, ComposedMetric, Aggregation
from mara_schema.sql_generation import database_identifier, table_alias_for_path



def date_dimension_xml(attribute: Attribute, name):
    """
    Return the mondrian xml fragment for a date dimension

    Requires the time.day table from https://github.com/mara/mara-etl-tools/blob/master/etl_tools/create_time_dimensions/create_tables.sql
    """
    dim = etree.Element("Dimension", name=name, type="TimeDimension", description=attribute.description,
                        foreignKey=database_identifier(name) + '_fk')
    hierarchy = etree.SubElement(dim, "Hierarchy", allMemberName=f'All {name.lower()}s', hasAll="true", name="By month")
    etree.SubElement(hierarchy, "Table", schema="time", name="day")
    etree.SubElement(hierarchy, "Level", name="Year", column="year_id", type="Integer",
                     levelType="TimeYears", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Quarter", column="quarter_id", type="Integer",
                     levelType="TimeQuarters", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Month", column="month_id", type="Integer",
                     levelType="TimeMonths", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Day", column="day_id", type="Integer",
                     levelType="TimeDays", uniqueMembers="true")
    hierarchy = etree.SubElement(dim, "Hierarchy", allMemberName=f'All {name.lower()}s', hasAll="true", name="By week")
    etree.SubElement(hierarchy, "Table", schema="time", name="day")
    etree.SubElement(hierarchy, "Level", name="Year", column="iso_year_id", type="Integer",
                     levelType="TimeYears", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Week", column="week_id", type="Integer",
                     levelType="TimeWeeks", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Day", column="day_id", type="Integer",
                     levelType="TimeDays", uniqueMembers="true")
    return dim


def duration_dimension_xml(attribute: Attribute, name):
    """
    Return the mondrian xml fragment for a duration dimension

    Requires the time.duration table from https://github.com/mara/mara-etl-tools/blob/master/etl_tools/create_time_dimensions/create_tables.sql
    """

    dim = etree.Element("Dimension", name=name, type="StandardDimension", description=attribute.description,
                        foreignKey=database_identifier(name) + '_fk')
    hierarchy = etree.SubElement(dim, "Hierarchy", allMemberName=f'All {name.lower()}s', hasAll="true", name="By month")
    etree.SubElement(hierarchy, "Table", schema="time", name="duration")
    etree.SubElement(hierarchy, "Level", name="Days", column="days", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Months", column="months", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Half years", column="half_years", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Years", column="years", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    hierarchy = etree.SubElement(dim, "Hierarchy", allMemberName=f'All {name.lower()}s', hasAll="true", name="By week")
    etree.SubElement(hierarchy, "Table", schema="time", name="duration")
    etree.SubElement(hierarchy, "Level", name="Days", column="days", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Weeks", column="weeks", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Four weeks", column="four_weeks", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    etree.SubElement(hierarchy, "Level", name="Years", column="years", type="Integer",
                     levelType="Regular", uniqueMembers="true")
    return dim


def cube_xml(data_set: DataSet, schema_fact_table: str, table_name: str, personal_data: bool,
             high_cardinality_attributes: bool):
    cube = etree.Element("Cube", name=data_set.name, description=data_set.entity.description,
                         defaultMeasure=list(data_set.metrics.keys())[0])

    etree.SubElement(cube, "Table", schema=schema_fact_table, name=table_name)

    for path, attributes in data_set.connected_attributes(include_personal_data=personal_data).items():
        for name, attribute in attributes.items():
            if attribute.type != Type.ARRAY \
                    and (high_cardinality_attributes or
                         (not high_cardinality_attributes
                          and not attribute.high_cardinality)):
                if attribute.type == Type.DATE:
                    dim = date_dimension_xml(attribute, name)
                    cube.append(dim)
                elif attribute.type == Type.DURATION:
                    dim = duration_dimension_xml(attribute, name)
                    cube.append(dim)
                elif not path:
                    dim = etree.SubElement(cube, "Dimension", name=name, description=attribute.description)
                    hierarchy = etree.SubElement(dim, "Hierarchy", allMemberName='All ' + name, hasAll="true")
                    etree.SubElement(hierarchy, "Level", name=name, column=attribute.column_name, uniqueMembers="true")

                else:
                    entity = path[-1].target_entity
                    dim = etree.SubElement(cube, "Dimension", name=name, description=attribute.description,
                                           foreignKey=table_alias_for_path(path) + '_fk')
                    hierarchy = etree.SubElement(dim, "Hierarchy", allMemberName='All ' + name, hasAll="true",
                                                 primaryKey=entity.pk_column_name)

                    etree.SubElement(hierarchy, "Table", name=entity.table_name,
                                     schema=entity.schema_name)
                    etree.SubElement(hierarchy, "Level", name=name,
                                     column=attribute.column_name, uniqueMembers="true")

    for name, metric in data_set.metrics.items():
        if isinstance(metric, SimpleMetric):
            if metric.aggregation in [Aggregation.COUNT, Aggregation.DISTINCT_COUNT]:
                data_type = 'Integer'
            else:
                data_type = 'Numeric'
            etree.SubElement(cube, "Measure", name=name, description=metric.description,
                             column=database_identifier(metric.name), aggregator=metric.aggregation,
                             formatString=metric.number_format, datatype=data_type)

    for name, metric in data_set.metrics.items():
        if isinstance(metric, ComposedMetric):
            formula = metric.formula_template.format(
                *[f'[{metric.name}]' for metric in metric.parent_metrics]).replace('[', '[Measures].[')

            measure = etree.SubElement(cube, "CalculatedMember", name=name, dimension="Measures",
                                       description=metric.description)
            hierarchy = etree.SubElement(measure, "Formula")
            hierarchy.text = formula
            etree.SubElement(measure, "CalculatedMemberProperty", name="FORMAT_STRING",
                             value=metric.number_format)

    return cube


def write_mondrian_schema(file_name: pathlib.Path,
                          data_set_tables: typing.Dict[DataSet, typing.Tuple[str, str]],
                          personal_data: bool = False,
                          high_cardinality_attributes: bool = False):
    """Generate XML schema as a file.

    Args:
        file_name: The path of the Mondrian schema file
        data_set_tables: A dictionary with data sets as keys and tuples of strings (database schema + table) as values
            e.g. {DataSet 1: ('schema 1', 'table 1')}
        personal_data: Whether to include attributes that are marked as personal data
        high_cardinality_attributes: Whether to include attributes that are marked as have a high cardinality

    """
    root = etree.Element("Schema", name='Mondrian')

    for data_set, (fact_table_schema, table_name) in data_set_tables.items():
        root.append(cube_xml(data_set, fact_table_schema, table_name, personal_data, high_cardinality_attributes))

    tree = etree.ElementTree(root)
    tree.write(file_name.absolute().as_uri(), encoding='utf-8', pretty_print=True, xml_declaration=True)

    return True



"""ACL and navigation entries for Saiku"""

import flask
from mara_page import acl, navigation

# The flask blueprint that does
blueprint = flask.Blueprint('mara_mondrian', __name__, static_folder='static',
                            template_folder='templates', url_prefix='/mondrian')

# Defines an ACL resource (needs to be handled by the application)
acl_resource_saiku = acl.AclResource(name='Saiku')


@blueprint.before_app_first_request  # configuration needs to be loaded before we can access it
def _create_acl_resource_for_each_data_set():
    import mara_schema.config
    for data_set in mara_schema.config.data_sets():
        resource = acl.AclResource(name=data_set.name)
        acl_resource_saiku.add_child(resource)


def saiku_navigation_entry():
    return navigation.NavigationEntry(
        label='Saiku', icon='bar-chart', uri_fn=lambda: flask.url_for('mara_mondrian.saiku'),
        description='Pivoting & ad hoc analysis')


@blueprint.route('/saiku')
@acl.require_permission(acl_resource_saiku)
def saiku():
    from mara_mondrian import config

    return flask.redirect(config.mondrian_server_external_url())


@blueprint.route('/saiku/authorize', methods=['POST'])
def saiku_authorize():
    """
    Authorization endpoint for Saiku.

    See https://github.com/project-a/mondrian-server#authentication--acl
    """
    allowed_cubes = [resource.name
                     for resource, allowed
                     in acl.user_has_permissions(flask.request.form['username'], acl_resource_saiku.children)
                     if allowed]

    return flask.jsonify({'allowed': True if allowed_cubes else False, 'cubes': allowed_cubes})



from setuptools import setup, find_packages
import re

def get_long_description():
    with open('README.md') as f:
        return re.sub('!\[(.*?)\]\(docs/(.*?)\)', r'![\1](https://github.com/mara/mara-mondrian/raw/master/docs/\2)', f.read())

setup(
    name='mara-mondrian',
    version='2.0.0',

    description='A python integration for the Saiku ad hoc analysis tool',

    long_description=get_long_description(),
    long_description_content_type='text/markdown',

    url = 'https://github.com/mara/mara-mondrian',

    install_requires=[
        "flask>=0.12",
        "mara-page>=1.1.0",
        "mara-schema>=1.0.0",
        "lxml>=3.8.0",
        "requests"
    ],

    python_requires='>=3.6',

    packages=find_packages(),

    author='Mara contributors',
    license='MIT',

    entry_points={},
)
